;---------------------------------------------------------------------
;                      WAVPlayer source code
;                       Jesús Arias (2024)
;---------------------------------------------------------------------

;---------------------------------------------------------------------
; Programming Conventions:
; Strict:
; - Core revision 6 required 
; - R7 is the Stack Pointer
; - R6 holds the return address for subroutines
; - Full-descending Stack
; Loose guidelines (see specific function comments):
; - R0, R1, R2 are used for arguments / results
;              Values may change on subroutine calls
; - R3, R4, R5 are used for local variables
;              Values are preserved on subroutine calls
;---------------------------------------------------------------------
;---------------------------------------------------------------------
;----------- I/O MAP  -----------

IOBASE=		0x20
IRQEN=		0x20
PFLAGS=		0x21
UARTDAT=	0x22
GPOUT=      0x23
SPI16=      0x24
SPI8=       0x25
DAC1=       0x26
DAC2=       0x27

;--------- Interrupts --------
IRQ0EN=		1
;--------- PFLAGS register bits -----------
UARTDV=		1		; Data valid (UART RX)
UARTRDY=	2		; Transmitter ready (UART TX)
SPIBUSY=    0x10	; SPI busy
SAMPLECK=	0x20	; DAC clock (smpling clock)
;--------- GPOUT ----------
LEDM=       0x1F    ; LED mask
CSB=		0x40	; SD Card /CS
SPIFAST=	0x80	; SPI FAST (1) / SLOW (0)

;--------------------------------------
;----------- VECTORS -----------------
;-------------------------------------- 

		org 	0			; RESET
init:	ldpc	r7			; stack at the end of memory
		word	stack	
		jr		start

		org		4			; IRQ 0
		jr		irq0

;--------------------------------------
; Area 0x20 to 0x3F reserved for I/O
;--------------------------------------

;--------------------------------------
;----------- VARIABLES (near) ---------
;--------------------------------------
; Located below 0x100 in order to load their addresses with LDI

		org	0x40		
fatbas:		; FAT variables
storage_begin:	
		word 0
		word 0
fat1_begin:	
		word 0
		word 0
root_cluster:
		word 0
		word 0

last_cluster:
		word 0
		word 0
l2spc:	word 0		; log2(sectors_per_cluster)
spc:	word 0		; sectors_per_cluster

		; File variables
filebas:
filelen:
		word 0
		word 0
filecluster:
		word 0
		word 0
filesector:
		word 0
;bufix:	word 0	; for file streaming
;bufmax:	word 0	

; File Selector variables

fselbas:	
fsel:	word 0	; index of selection (<32)
pos:	word 0	; file position after offset
ppos:	word 0	; file position before offset
offset:	word 0	; first file displayed
curdir:	word 0 	; current directory cluster
		word 0
scr:	word 0	; scratch data

; Audio stream
auplay:	word 0	; flag play / silence
auix:	word 0	; read index

;--------------------------------------
;-----------   Interrupts -------------
;-------------------------------------- 

;		org		0x80
irq0:	subi	r7,3
		st		(r7+0),r0
		st		(r7+1),r1
		st		(r7+2),r2

		ldi		r1,auplay
		ld		r0,(r1)
		jnz		ir1
		ldi		r2,IOBASE
		st		(r2+DAC1-IOBASE),r0
		st		(r2+DAC2-IOBASE),r0
		jr		ir9
		
ir1:	ld		r0,(r1+1)	; index
		or		r2,r0,r0
		addi	r0,1
		andi	r0,0xff
		st		(r1+1),r0
		add		r0,r2,r2	; ix*2
		ldpc	r1
		word	audio
		add		r1,r1,r0
		ld		r0,(r1)
		ldi		r2,IOBASE
		rori	r0,r0,8
		st		(r2+DAC1-IOBASE),r0
		ld		r0,(r1+1)
		rori	r0,r0,8
		st		(r2+DAC2-IOBASE),r0

ir9:	ld		r0,(r7+0)			; return from interrupt
		ld		r1,(r7+1)
		ld		r2,(r7+2)
		addi	r7,3
		reti


;-----------------------------------------------
;       SPI
;-----------------------------------------------
;	SPI transfers (16 / 8 bits)
; Parameters:	R0: data to send
; Returns:		R0: data received
; Modiffies:	no other register

SPIxfer16:
		subi	r7,1
		st		(r7+0),r1
		ldi		r1,IOBASE
		st		(r1+SPI16-IOBASE),r0
		jr		bspx1
SPIxfer:subi	r7,1
		st		(r7+0),r1
		ldi		r1,IOBASE
		st		(r1+SPI8-IOBASE),r0
bspx1:	ld		r0,(r1+PFLAGS-IOBASE)
		andi	r0,SPIBUSY
		jnz		bspx1
		ld		r0,(r1+SPI16-IOBASE)
		ld		r1,(r7+0)
		addi	r7,1
		jind	r6			; retornamos	
		
;--------------------------------------
;-----------   MAIN   -----------------
;-------------------------------------- 

start:
		ldi		r0,CSB|0x10		; SD_CS = H, LED[4] on
		ldi		r1,IOBASE
		st		(r1+GPOUT-IOBASE),r0
		ldi		r0,1
		st		(r1+IRQEN-IOBASE),r0
		
;--------------------------------------------------
;--------------------------------------
;  		File selection loop
;--------------------------------------
;--------------------------------------------------

filesel:
		; card init
		jal		sd_init
		or		r0,r0,r0
		jnz		inie1
		jal		FAT_init
		or		r0,r0,r0
		jnz		inie2
		jr		fs000
inie1:	ldpc	r0
		word	2*txterrsd
		jr		inie3
inie2:	ldpc	r0
		word	2*txterrfat
inie3:	jal		puts
		jr		filesel

fs000:	; display init
		;jal		frameit
		ldpc	r0
		word	2*txto1
		jal		puts

		; start at root directory
		ldi		r2,fselbas
		ldi		r1,fatbas
		ld		r0,(r1+root_cluster-fatbas)
		st		(r2+curdir-fselbas),r0
		ld		r0,(r1+root_cluster-fatbas+1)
		st		(r2+curdir-fselbas+1),r0

fs00:	ldi		r2,fselbas
		ldi		r0,0	; fsel=pos=ppos=offset=0
		st		(r2+fsel-fselbas),r0
		st		(r2+pos-fselbas),r0
		st		(r2+ppos-fselbas),r0
		st		(r2+offset-fselbas),r0

fs0:	; read current directory
		
		ldi		r2,fselbas
		ldi		r1,fatbas
		ld		r0,(r2+curdir-fselbas)
		st		(r1+filecluster-fatbas),r0
		ld		r0,(r2+curdir-fselbas+1)
		st		(r1+filecluster-fatbas+1),r0
		ldi		r0,0
		st		(r1+filesector-fatbas),r0
		subi	r0,1
		st		(r1+filelen-fatbas),r0
		st		(r1+filelen-fatbas+1),r0

fs1:	; scan directory entries
		ldpc	r3
		word	fatbuf
		jal		rdfile
		ldpc	r1
		word	2*fatbuf

		ldi		r4,16
fs2:	jal		rdbyte
		or		r0,r0,r0
		jz		fs99
		cmpi	r0,0xE5		; deleted?
		jz		fsskip
		addi	r1,10
		jal		rdbyte
		subi	r1,12
		ldi		r6,0x30		; file or subdir
		and		r6,r0,r6
		jz		fsskip
		; File OK: count it
		ldi		r6,fselbas	; subdir attribute in scr
		andi	r0,0x10
		st		(r6+scr-fselbas),r0
		ld		r0,(r6+ppos-fselbas)
		ld		r2,(r6+offset-fselbas)
		sub		r2,r0,r2
		jz		fss1
		addi	r0,1
		st		(r6+ppos-fselbas),r0
		jr		fsskip
fss1:	; copy file metadata
		ld		r0,(r6+pos-fselbas)
		rori	r0,r0,12	; pos*16
		ldpc	r2
		word	filedata
		add		r6,r0,r2
		shr		r2,r1
		ldi		r5,16
fss11:	ld		r0,(r2)
		st		(r6),r0
		addi	r6,1
		addi	r2,1
		subi	r5,1
		jnz		fss11
		; print name
		ldi		r0,'\e'
		jal		putch
		ldi		r0,'['
		jal		putch
		ldi		r6,fselbas
		ld		r0,(r6+pos-fselbas)
		ld		r5,(r6+fsel-fselbas)
		sub		r5,r0,r5
		jnz		fss12
		ldi		r0,'7'
		jal		putch
fss12:	ldi		r0,'m'
		jal		putch

		jal		gotopos
		ldi		r0,8
		jal		bytenpt
		ldi		r0,' '
		jal		putch
		ldi		r0,3
		jal		bytenpt

		ldi		r0,' '
		jal		putch
		
		ldi		r6,fselbas
		ld		r5,(r6+scr-fselbas)
		or		r5,r5,r5
		jz		fss13
		ldi		r0,'D'
fss13:	jal		putch
		subi	r1,11
		
fss2:
		ldi		r6,fselbas
		ld		r0,(r6+pos-fselbas)
		addi	r0,1
		st		(r6+pos-fselbas),r0
		cmpi	r0,32
		jz		fs99

fsskip:		; Next file entry
		addi	r1,32
		subi	r4,1
		jnz		fs2
		jr		fs1	; Read more sectors of directory

fs99:	; clear until the end of screen
		ldi		r5,fselbas
		ld		r0,(r5+pos-fselbas)
		or		r4,r0,r0
fs9900:	cmpi	r0,32
		jz		fs9903
		jal		gotopos
		ldi		r0,'\e'
		jal		putch
		ldi		r0,'['
		jal		putch
		ldi		r0,'K'
		jal		putch

		ld		r0,(r5+pos-fselbas)
		addi	r0,1
		st		(r5+pos-fselbas),r0
		jr		fs9900

fs9903:	st		(r5+pos-fselbas),r4
		;------------------------------------------------
		;-------------- Wait for KEY action -------------
		;------------------------------------------------
wkey:	jal		getch
		cmpi	r0,'B'
		jz		kdown
		cmpi	r0,'A'
		jz		kup
		cmpi	r0,'C'
		jz		kright
		cmpi	r0,'D'
		jz		kleft
		
		cmpi	r0,'\n'		;enter
		jz		kenterspc
		cmpi	r0,' '		;space
		jz		kenterspc

		jr		wkey

		;------------------------------------------------
		;-------------- KEY actions -------------
		;------------------------------------------------

kup:	ld		r0,(r5+fsel-fselbas)
		jz		fs999
		subi	r0,1
		st		(r5+fsel-fselbas),r0
		jr		fs999

kdown:	ld		r0,(r5+fsel-fselbas)
		addi	r0,1
		st		(r5+fsel-fselbas),r0
		ld		r6,(r5+pos-fselbas)
		sub		r0,r0,r6
		jnc		fs999
		subi	r6,1
		st		(r5+fsel-fselbas),r6
		jr		fs999

kright:	ld		r0,(r5+fsel-fselbas)
		addi	r0,16
		ld		r1,(r5+pos-fselbas)
		sub		r1,r0,r1
		jc		fs9902
		st		(r5+fsel-fselbas),r0
		jr		fs999
fs9902:	ld		r1,(r5+pos-fselbas)
		cmpi	r1,32
		jmi		fs999
		ld		r1,(r5+offset-fselbas)
		addi	r1,16
		st		(r5+offset-fselbas),r1
		jr		fs999

kleft:	ld		r0,(r5+fsel-fselbas)
		cmpi	r0,16
		jnc		fs9912
		subi	r0,16
		st		(r5+fsel-fselbas),r0
		jr		fs999

fs9912:	ld		r0,(r5+offset-fselbas)
		jz		fs999
		subi	r0,16
		st		(r5+offset-fselbas),r0
		jr		fs999


kenterspc:	; File or subdir selected
		or		r2,r0,r0
		ld		r0,(r5+fsel-fselbas)
		rori	r0,r0,12	; fsel*16
		ldpc	r1
		word	filedata
		add		r1,r0,r1	; r1: file info
		or		r4,r1,r1
		ld		r0,(r1+5)
		ldi		r6,0x10		; Subdirectory attributte
		and		r6,r0,r6
		jz		fs995
		; Subdirectory
		ld		r0,(r1+13)	; cluster LSB
		ld		r1,(r1+10)	; cluster MSB
		rori	r0,r0,8
		rori	r1,r1,8
		or		r6,r0,r1
		jz		filesel		; root cluster
		st		(r5+curdir-fselbas),r0
		st		(r5+curdir-fselbas+1),r1
		ldi		r6,filelen	; filelen = max.
		ldpc	r0
		word	0xffff
		st		(r6),r0
		st		(r6+1),r0
		jr		fs00

fs995:	ld		r0,(r1+4)	; file extension
		ldpc	r6
		word	0x5741		; "WA"
		sub		r6,r0,r6
		jnz		fs999
		; File WAV
		ld		r0,(r1+13)	; cluster LSB
		ld		r2,(r1+10)	; cluster MSB
		rori	r0,r0,8
		rori	r2,r2,8
		ldi		r6,filebas
		st		(r6+filecluster-filebas),r0
		st		(r6+filecluster-filebas+1),r2
		ld		r0,(r1+14)	; filelen LSB
		ld		r2,(r1+15)	; filelen MSB
		rori	r0,r0,8
		rori	r2,r2,8
		st		(r6+filelen-filebas),r0
		st		(r6+filelen-filebas+1),r2		
		ldi		r0,0
		st		(r6+filesector-filebas),r0
		jal		player

fs999:	ldi		r0,0
		ldi		r5,fselbas
		st		(r5+pos-fselbas),r0
		st		(r5+ppos-fselbas),r0
		jr		fs0
		
;--------------------------------------
;  		SD card functions
;--------------------------------------
; CS high/Low
; Parameters:	none
; Returns:		none
; Modiffies:	no register
sd_cs_h:
		subi	r7,2
		st		(r7+0),r0
		st		(r7+1),r6
		ldi		r6,IOBASE
		ld		r0,(r6+GPOUT-IOBASE)
		ori		r0,CSB
		jr		sd_cs1
		
sd_cs_l:
		subi	r7,2
		st		(r7+0),r0
		st		(r7+1),r6
		ldi		r6,IOBASE
		ld		r0,(r6+GPOUT-IOBASE)
		andi	r0,0xFF^CSB
sd_cs1:	st		(r6+GPOUT-IOBASE),r0
		ld		r0,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;--------------------------------------
; SD command
;
; Parameters:	R0: command
;				R2:R1 argument (32 bits)
; Returns:		R0: r1 state
; Modiffies:	nothing else

sd_cmd:
		subi	r7,3
		st		(r7+0),r1
		st		(r7+1),r4
		st		(r7+2),r6
	
		jal		sd_cs_l
		ldi		r4,0x87		; CRC
		or		r0,r0,r0
		jnz		sdcm1
		ldi		r4,0x95		; CRC
sdcm1:	ldi		r6,0x40
		or		r0,r0,r6
		jal		SPIxfer
		or		r0,r2,r2	; arg MSBs
		jal		SPIxfer16
		or		r0,r1,r1	; arg LSBs
		jal		SPIxfer16	
		or		r0,r4,r4	; CRC
		jal		SPIxfer

		ldi		r1,255
		;ldpc	r1
		;word	0x7000
sdcm2:	ldi		r0,0xff
		jal	SPIxfer
		cmpi	r0,0xff
		jnz		sdcm3
		subi	r1,1
		jpl		sdcm2

sdcm3:	or		r4,r0,r0
retcmd:	ldi		r0,0xff
		jal		SPIxfer
		jal		sd_cs_h
		ldi		r0,0xff
		jal		SPIxfer
		or		r0,r4,r4
		
		ld		r1,(r7+0)
		ld		r4,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6

;--------------------------------------
; SD command, 32 bit result
;
; Parameters:	R0: command
;				R2:R1 argument (32 bits)
; Returns:		R0: r1-state
;				R2:R1 32 bit result (sometimes)
; Modiffies:	nothing else
sd_cmd32:
		subi	r7,3
		st		(r7+0),r1
		st		(r7+1),r4
		st		(r7+2),r6
	
		jal		sd_cs_l
		ldi		r4,0x87		; CRC
		or		r0,r0,r0
		jnz		sdcm31
		ldi		r4,0x95		; CRC
sdcm31:	ori		r0,0x40
		jal		SPIxfer
		or		r0,r2,r2	; arg MSBs
		jal		SPIxfer16
		or		r0,r1,r1	; arg LSBs
		jal		SPIxfer16	
		or		r0,r4,r4	; CRC
		jal		SPIxfer

		ldi		r1,255
sdcm32:	ldi		r0,0xff
		jal		SPIxfer
		cmpi	r0,0xff
		jnz		sdcm33
		subi	r1,1
		jpl		sdcm32

sdcm33:
		or		r4,r0,r0
		ldpc	r0
		word	0xffff
		jal		SPIxfer16
		or		r2,r0,r0
		ldpc	r0
		word	0xffff
		jal		SPIxfer16
		or		r1,r0,r0
		jr		retcmd

;--------------------------------------
; SD init
;
; Parameters:	none
; Returns:		R0: 0 OK
; Modiffies:	R1,R2

sd_init:
		subi	r7,2
		st		(r7+0),r3
		st		(r7+1),r6
		ldi		r1,IOBASE			; slow clock and CS high
		ld		r0,(r1+GPOUT-IOBASE)
		andi	r0,0xFF^SPIFAST
		ori		r0,CSB
		st		(r1+GPOUT-IOBASE),r0

		ldi		r3,10				; 80 clock pulses
sdini1:	ldi		r0,0xff
		jal		SPIxfer
		subi	r3,1
		jnz		sdini1

		jal		sd_cs_l

		ldi		r3,128
sdini2:	ldi		r0,0	; CMD go idle state
		ldi		r2,0	; arg MSB
		ldi		r1,0	; arg LSB
		jal		sd_cmd
		cmpi	r0,1
		jz		sdini3
		subi	r3,1
		jnz		sdini2
		ldi		r0,1	; Error:timeout
sdini99:
		ld		r3,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
		
sdini3:	ldi		r3,255
sdini35:
		ldi		r0,8	; SEND_IF_COND
		ldi		r2,0
		ldpc	r1
		word	0x01aa
		jal		sd_cmd32
		cmpi	r0,1
		jz		sdini40
		subi	r3,1
		jpl		sdini35

sdini40:
		ldi		r3,255
sdini4:	ldi		r0,55	; APP_CMD
		ldi		r2,0
		ldi		r1,0
		jal		sd_cmd
		ldi		r0,41	; SD_SEND_OP_COND,0x40000000
		ldpc	r2
		word	0x4000
		ldi		r1,0
		jal		sd_cmd
		cmpi	r0,0
		jz		sdini5
		subi	r3,1
		jpl		sdini4
		ldi		r0,2	; Error card initialization failed
		jr		sdini99

sdini5:	ldi		r3,255
sdini6:	ldi		r0,58	; Read OCR
		ldi		r2,0
		ldi		r1,0
		jal		sd_cmd32
		cmpi	r0,0
		jz		sdini9
		subi	r3,1
		jpl		sdini6
		ldi		r0,3	; Error read OCR
		jr		sdini99

sdini9:	ldi		r1,IOBASE
		ld		r0,(r1+GPOUT-IOBASE)
		ori		r0,SPIFAST
		st		(r1+GPOUT-IOBASE),r0
		ldi		r0,0
		jr		sdini99

;--------------------------------------
; SD read block
;
; Parameters:	R2:R1 :	nº sector
;				R3: 	destination address
; Returns:		R0: 0 OK, !=0 Error
; Modiffies:	R3 += 256

sd_rd_block:
		subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6
		ldi		r0,17	; READ SINGLE BLOCK
		jal		sd_cmd
		or		r0,r0,r0
		jnz		sdrd9

		jal		sd_cs_l
		ldpc	r1		; timeout
		word	0x300
sdrd2:	ldi		r0,0xff
		jal		SPIxfer
		cmpi	r0,0xFE
		jz		sdrd3
		subi	r1,1
		jnz		sdrd2
		ldi		r0,4
		jr		sdrd9

sdrd3:	ldi		r1,255	; 256 words
sdrd4:	ldpc	r0
		word	0xffff
		jal		SPIxfer16
		st		(r3),r0
		addi	r3,1
		subi	r1,1
		jpl		sdrd4

		ldi		r0,0xff	; discard results and CRC
		jal		SPIxfer
		ldi		r0,0xff
		jal		SPIxfer
		ldi		r0,0xff
		jal		SPIxfer
		ldi		r0,0	; OK

sdrd9:	jal		sd_cs_h
		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6


;--------------------------------------
; SD write block
;
; Parameters:	R2:R1 :	nº sector
;				R3: 	source address
; Returns:		R0: 0 OK, !=0 Error
; Modiffies:	R3 += 256	
sd_wr_block:
		subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6
		ldi		r0,24	; WRITE  SINGLE BLOCK
		jal		sd_cmd
		or		r0,r0,r0
		jnz		sdrd9

		jal		sd_cs_l		
		ldi		r0,0xFE	;Send start block token 0xfe
		jal		SPIxfer

		ldi		r1,255	; Send data
sdwr4:	ld		r0,(r3)
		addi	r3,1
		jal		SPIxfer16
		subi	r1,1
		jpl		sdwr4

		or		r0,r1,r1
		jal		SPIxfer16
	
		ldi		r1,0	; wait for data response
sdwr5:	ldi		r0,0xff
		jal		SPIxfer
		cmpi	r0,0xff
		jnz		sdwr6
		subi	r1,1
		jnz		sdwr5
		jr		sdwr91		; timeout
sdwr6:	andi	r0,0x1F		; AAA='010' - data accepted ;
		cmpi	r0,5	
		jnz		sdwr91		; error	

		ldi		r1,0	;Wait while card busy
sdwr7:	ldi		r0,0xff
		jal		SPIxfer
		or		r0,r0,r0
		jnz		sdwr92
		subi	r1,1
		jnz		sdwr7

sdwr91:	ldi		r1,1	; return error
sdwr93:	ldi		r0,0xff
		jal		SPIxfer
		jal		sd_cs_h
		jal		SPIxfer
		or		r0,r1,r1
		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
sdwr92:	ldi		r1,0	; return OK
		jr		sdwr93

;--------------------------------------
;--------------------------------------
; FAT init
; Parameters:	none
; Returns:	A: 0 OK, !=0 Error
; Modiffies:	lots

FAT_init:
		subi	r7,3
		st		(r7+0),r4
		st		(r7+1),r5
		st		(r7+2),r6
		ldi		r5,fatbas
		ldi		r2,0
		ldi		r1,0
		ldpc	r3
		word	fatbuf
		jal		sd_rd_block	; read partition table
		or		r0,r0,r0
		jnz		fini99
		subi	r3,1
		ld		r0,(r3)
		rori	r0,r0,8
		ldpc	r6
		word	0xAA55
		sub		r6,r0,r6
		jnz		fini99
		ldi		r6,28
		sub		r3,r3,r6
		ld		r1,(r3)		; starting sector of partition #1
		ld		r2,(r3+1)
		rori	r1,r1,8
		rori	r2,r2,8
		subi	r3,227
		jal		sd_rd_block	; superblock
		or		r0,r0,r0
		jnz		fini99

		ldpc	r6			; Sectors/Cluster
		word	0x100
		sub		r6,r3,r6
		ld		r0,(r6+6)
		andi	r0,0xff
		st		(r5+spc-fatbas),r0
		ldi		r6,0
fini1:	addi	r6,1
		shr		r0,r0
		jnz		fini1
		subi	r6,1
		st		(r5+l2spc-fatbas),r6
	
		ldpc	r6			
		word	0x100
		sub		r3,r3,r6
		ld		r0,(r3+7)	; reserved sectors
		rori	r0,r0,8
		add		r1,r1,r0
		adci	r2,0
		st		(r5+fat1_begin-fatbas),r1
		st		(r5+fat1_begin-fatbas+1),r2
	
		ld		r0,(r3+8)	; number of FATs
		rori	r0,r0,8
		ldi		r6,0x12
		add		r3,r3,r6
		ld		r4,(r3)		;sectors/FAT
		ld		r6,(r3+1)
		rori	r4,r4,8
		rori	r6,r6,8
fini3:	add		r1,r1,r4
		adc		r2,r2,r6
		subi	r0,1
		jnz		fini3
		st		(r5+storage_begin-fatbas),r1
		st		(r5+storage_begin-fatbas+1),r2

		ld		r4,(r3+4)
		ld		r6,(r3+5)
		rori	r4,r4,8
		rori	r6,r6,8
		st		(r5+root_cluster-fatbas),r4
		st		(r5+root_cluster-fatbas+1),r6

		ldpc	r6
		word	0xffff
		st		(r5+last_cluster-fatbas),r6
		st		(r5+last_cluster-fatbas+1),r6
		
fini99:	ld		r4,(r7+0)
		ld		r5,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6
		
;--------------------------------------
;--------------------------------------
; next_cluster
; Parameters:	R2:R1 : current cluster
; Returns:		R0 : 0 if no error
;				R2:R1 : next cluster
; Modiffies:	nothing else

next_cluster:
		subi	r7,4
		st		(r7+0),r3
		st		(r7+1),r4
		st		(r7+2),r5
		st		(r7+3),r6
		; 	if ((cluster^FATvar.last_cluster)&0xffffff80) {
		ldi		r3,fatbas
		ld		r4,(r3+last_cluster-fatbas)
		ld		r5,(r3+last_cluster-fatbas+1)
		xor		r5,r5,r2
		xor		r4,r4,r1
		ldpc	r6
		word	0xff80
		and		r4,r4,r6
		or		r4,r4,r5
		jz		nxc1
; 		// read new FAT chunk
; 		FATvar.last_cluster=cluster;
		st		(r3+last_cluster-fatbas),r1
		st		(r3+last_cluster-fatbas+1),r2
; 		if (rdsector(FATvar.fat1_begin+(cluster>>7),FATbuf)) return 0;
		; (cluster>>7)
		ldi		r6,7
nxc2:	shr		r2,r2
		rorc	r1,r1
		subi	r6,1
		jnz		nxc2
		; +fat1_begin
		ld		r4,(r3+fat1_begin-fatbas)
		ld		r5,(r3+fat1_begin-fatbas+1)
		add		r1,r1,r4
		adc		r2,r2,r5
		ldpc	r3
		word	fatbuf
		jal		sd_rd_block
		or		r0,r0,r0
		jnz		nxc99
		ldi		r3,fatbas
		ld		r1,(r3+last_cluster-fatbas)
; 	pi=(u32 *)FATbuf; 	return pi[cluster&127];		
nxc1:	ldi		r5,127
		and		r5,r1,r5
		add		r5,r5,r5
		ldpc	r6
		word	fatbuf	; byte swapped area
		add		r5,r5,r6
		ld		r1,(r5)
		ld		r2,(r5+1)
		rori	r1,r1,8
		rori	r2,r2,8
		
nxc99:	ld		r3,(r7+0)
		ld		r4,(r7+1)
		ld		r5,(r7+2)
		ld		r6,(r7+3)
		addi	r7,4
		jind	r6

;--------------------------------------
;--------------------------------------
; rdfile
; Parameters:	R3 : pointer to buffer
; Returns:		R0 : bytes read (512 unless EOF)
;				R3+= 256
; Modiffies:	R1, R2

rdfile:	subi	r7,5
		st		(r7+0),r3
		st		(r7+1),r4
		st		(r7+2),r5
		st		(r7+3),r6

		ldi		r5,fatbas
;	// next sector
;	nc=FATvar.filecluster;
;	ns=FATvar.filesector+1;
;	i=1<<FATvar.l2spc;	// i = sectors per cluster
		ld		r1,(r5+filecluster-fatbas)
		ld		r2,(r5+filecluster-fatbas+1)
		ld		r4,(r5+filesector-fatbas)
		addi	r4,1
		ld		r6,(r5+spc-fatbas)
;if (ns>=spc) {ns=0; nc=next_cluster32(nc);}
		sub		r6,r4,r6
		jnc		rdf1
		ldi		r4,0
		jal		next_cluster
;	if(FATvar.filecluster>=0x0ffffff0) return 0; //EOF
rdf1:	st		(r7+4),r4	; save ns
		ldpc	r0
		word	0x0FFF
		ld		r6,(r5+filecluster-fatbas+1)
		sub		r0,r0,r6
		jnz		rdf2
		ldpc	r0
		word	0xFFF0
		ld		r6,(r5+filecluster-fatbas)
		and		r6,r6,r0
		sub		r0,r6,r0
		jnz		rdf2
		ldi		r0,0
		jr		rdf99
; if (buf==FATbuf) FATvar.last_cluster=-1; 		
rdf2:	ldpc	r0
		word	fatbuf
		sub		r0,r3,r0
		jnz		rdf3
		ldpc	r0
		word	0xFFFF
		st		(r5+last_cluster-fatbas),r0
		st		(r5+last_cluster-fatbas+1),r0
;	i=(	FATvar.storage_begin+((FATvar.filecluster-2)<<FATvar.l2spc) ) + FATvar.filesector;
rdf3:	ld		r4,(r5+filecluster-fatbas)
		ld		r6,(r5+filecluster-fatbas+1)
		subi	r4,2
		sbci	r6,0
		ld		r0,(r5+l2spc-fatbas)
rdf4:	add		r4,r4,r4
		adc		r6,r6,r6
		subi	r0,1
		jnz		rdf4
		ld		r0,(r5+storage_begin-fatbas)
		add		r4,r4,r0
		ld		r0,(r5+storage_begin-fatbas+1)
		adc		r6,r6,r0
		ld		r0,(r5+filesector-fatbas)
		add		r4,r4,r0
		adci	r6,0
;	FATvar.filesector=ns;
;	FATvar.filecluster=nc;
		ld		r0,(r7+4)
		st		(r5+filesector-fatbas),r0
		st		(r5+filecluster-fatbas),r1
		st		(r5+filecluster-fatbas+1),r2
;	if (rdsector(i,buf)) return 0;
		or		r1,r4,r4
		or		r2,r6,r6
		jal		sd_rd_block
		or		r0,r0,r0
		jnz		rdf99
;	if(FATvar.filelen>512) {FATvar.filelen-=512; return 512;}
;	else {i=FATvar.filelen; FATvar.filelen=0; return i; }
		ldpc	r0
		word	512
		ld		r4,(r5+filelen-fatbas)
		sub		r4,r4,r0
		ld		r6,(r5+filelen-fatbas+1)
		sbci	r6,0
		jpl		rdf5
		ld		r0,(r5+filelen-fatbas)
		ldi		r4,0
		ldi		r6,0
rdf5:	st		(r5+filelen-fatbas),r4
		st		(r5+filelen-fatbas+1),r6
		
rdf99:	ld		r3,(r7+0)
		ld		r4,(r7+1)
		ld		r5,(r7+2)
		ld		r6,(r7+3)
		addi	r7,5
		jind	r6	

;-----------------------------------------------
;       read a byte 
;-----------------------------------------------
; Parameters:	R1: byte address (=mem_address *2 +odd/even)
; Returns:		R0: byte read
; Modiffies:	R1++
; Notes:		Big endian memory assumed

rdbyte:	shr		r0,r1
		ld		r0,(r0)
		jc		rdb1
		rori	r0,r0,8
rdb1:	andi	r0,0xff
		addi	r1,1
		jind	r6

;-----------------------------------------------
;       bytenpt
;-----------------------------------------------
; Parameters:	R0: number of bytes to print
;				R1: source byte pointer
; Returns:		none
; Modiffies:	R1+=R0
bytenpt:
		subi	r7,3
		st		(r7+0),r0
		st		(r7+1),r5
		st		(r7+2),r6
		or		r5,r0,r0

bycpr1:	jal		rdbyte	
		jal		putch
		subi	r5,1
		jnz		bycpr1
	
		ld		r0,(r7+0)
		ld		r5,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6

		
txterrsd:	asczbe "Error:  SDinit\n"
txterrfat:	asczbe "Error: FATinit\n"
txto1:		asczbe "\e[2J\e[H\t\tWAVplay J. Arias (2023)"
txto2:		asczbe "\e[4;1H\e[J\e[m"
txto3:		asczbe "Playing... <enter stops>"

endtxt:

;----------------------------------------------
; Send data to UART
; Parameters:	R0: data to send
; Returns:		none
; Modiffies:	none

putch:	subi	r7,1
		st		(r7),r6
ptch1:	ldi		r6,IOBASE		; wait while TX busy
		ld		r6,(r6+PFLAGS-IOBASE)		
		andi	r6,UARTRDY
		jz		ptch1
		ldi		r6,IOBASE		
		st		(r6+UARTDAT-IOBASE),r0	; send data
		ld		r6,(r7)
		addi	r7,1
		jind	r6		

;----------------------------------------------
; Get data from UART (blocking)
; Parameters:	none
; Returns:		R0: data received
; Modiffies:	none

getch:	ldi		r0,IOBASE	
		ld		r0,(r0+PFLAGS-IOBASE)	; wait until data valid
		andi	r0,UARTDV
		jz		getch
		ldi		r0,IOBASE
		ld		r0,(r0+UARTDAT-IOBASE)	; get the data
		jind	r6

;-----------------------------------------------
;       puts 
;-----------------------------------------------
; Parameters:	R0: byte pointer
; Returns:		R0:	byte pointer updated R0+=strlen(string)
; Modiffies:	none
;
puts:	subi	r7,2
		st		(r7+0),r1
		st		(r7+1),r6
		or		r1,r0,r0
pts1:	jal		rdbyte	
		or		r0,r0,r0
		jz		pts2	
		jal		putch
		jr		pts1
pts2:	or		r0,r1,r1
		ld		r1,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6


;-----------------------------------------------
;		print <esc>[Hn,m depending on pos
;-----------------------------------------------
; Parameters:	pos: file selected
; Returns:		R2:	 byte pointer to video
; Modiffies:	none
gotopos:
		subi	r7,3
		st		(r7+0),r0
		st		(r7+1),r1
		st		(r7+2),r6

		ldi		r0,'\e'
		jal		putch
		ldi		r0,'['
		jal		putch

		ldi		r1,pos
		ld		r0,(r1)
		andi	r0,15
		addi	r0,4
		or		r1,r0,r0
		cmpi	r0,10
		jmi		gtp4
		ldi		r0,'1'
		jal		putch
		or		r0,r1,r1
		subi	r0,10
gtp4:	addi	r0,'0'
		jal		putch
		ldi		r0,';'
		jal		putch

		ldi		r1,pos
		ld		r0,(r1)
		rori	r0,r0,4
		andi	r0,3
		jz		gtp5
		add		r0,r0,r0
		addi	r0,'0'
		jal		putch
gtp5:	ldi		r0,'1'
		jal		putch
		ldi		r0,'H'
		jal		putch

		ld		r0,(r7+0)
		ld		r1,(r7+1)
		ld		r6,(r7+2)
		addi	r7,3
		jind	r6	

;--------------------------------------------------
;--------------------------------------
;  		Player
;--------------------------------------
;--------------------------------------------------
player:	subi	r7,2
		st		(r7+0),r3
		st		(r7+1),r6
		
		ldpc	r0
		word	2*txto2
		jal		puts
		ldpc	r0
		word	2*txto3
		jal		puts
		
		ldpc	r3		; preload buffers
		word	audio
		jal		rdfile
		ldpc	r3		; buffer1
		word	audio+256
		jal		rdfile
		
		ldi		r1,auplay
		ldi		r0,32
		st		(r1+1),r0	; index
		st		(r1),r0		; !=0 => play
						
plb1:	ldi		r1,auplay
		ld		r0,(r1+1)
		andi	r0,0x80
		jz		plb1
		ldpc	r3		; buffer1
		word	audio
		jal		rdfile
		ori		r0,0
		jz		ple9
		
plb2:	ldi		r1,auplay
		ld		r0,(r1+1)
		andi	r0,0x80
		jnz		plb2
		ldpc	r3		; buffer2
		word	audio+256
		jal		rdfile	
		ori		r0,0
		jz		ple9
		ldi		r1,IOBASE
		ld		r0,(r1+PFLAGS-IOBASE)
		andi	r0,UARTDV
		jz		plb1
		ld		r0,(r1+UARTDAT-IOBASE)
		cmpi	r0,'\n'
		jz		ple9
		jr		plb1

ple9:	ldi		r1,auplay
		ldi		r0,0
		st		(r1),r0		; ==0 => silence
		
		ldpc	r0
		word	2*txto2
		jal		puts
		
		ld		r3,(r7+0)
		ld		r6,(r7+1)
		addi	r7,2
		jind	r6
;-------- End of code
endcode:

;-------- special areas at memory top --------
fatbuf:		;block 512 bytes ; part of the File Allocation Table
filedata=	fatbuf+256	;block 32*32 bytes ; copy of "decent" files
audio=		filedata

enddata=	audio+512
stack=  	2048
;-------- end of memory -------

