//----------------------------------------------------------------------------
//-- 2nd-order Sigma-Delta DAC
//   J. Arias (2023) Public domain source
//----------------------------------------------------------------------------

module sd_DAC(
	input clk,		// around 3MHz, (2.82MHz for 44100/2^n)
	input [11:0]in,	// 12 bits, signed
	output out,		// output bitstream
	output sclk		// sampling clock (clk/OSR)
);

parameter OSR=64;	// Oversampling Ratio
localparam NDIV=$clog2(OSR);

// Clock divider for sampling rate generation
reg [NDIV-1:0]sdiv=0;
always @(posedge clk) sdiv<=sdiv+1;
assign sclk=sdiv[NDIV-1];

reg [11:0]inreg;
always @(negedge sclk) inreg<=in;


////////////////////////////////////////////////////////////
// optional upsampling by 4. (for 22 and 11 KSamples/s)
// (sign extension when required)
`ifdef INTERPOLATOR

	wire s4clk=sdiv[NDIV-3];

	//// Interpolator: fsampling x 4 
	reg [15:0]x1=0;
	reg [15:0]x2=0;
	reg [15:0]i1=0;
	reg [15:0]i2=0;
	reg [15:0]xm=0; 

	always @(posedge s4clk) begin
		i2<=i1;
		i1<={i1[14:0],1'b0} - i2 + xm;
		if (sdiv[NDIV-1]&sdiv[NDIV-2]) begin
			x2<=x1;
			x1<={{4{inreg[11]}},inreg};
			xm<={{4{inreg[11]}},inreg} - {x1[14:0],1'b0} + x2;
		end
	end
	// The SD modulator input, x, comes from the 12 MSB bits of the interpolator
	// (sign extended to 15 bits)
	wire [14:0]x={{4{i1[15]}},i1[14:4]};
`else
    // The SD modulator input, x, comes directly from input (with sign extension)
	wire [14:0]x={{3{inreg[11]}},inreg};
`endif

///////////////////// SD MODULATOR (DAC) ///////////////////////////
// 15 bits for variables in order to avoid overflows
wire [14:0]err;		// Quantization error
reg  [14:0]e1=0;	// error 1 cycle before
reg  [14:0]e2=0;	// error 2 cycle before
wire [14:0]y;		// filter output
wire outp;			// output bit
reg out=0;			// registerred output (to avoid glitches & variable delays)

assign y=x+{e1[13:0],1'b0}+(~e2)+1'b1;	// y=x+2*e1-e2 
assign outp=y[14];						// output = sign bit
assign err=y+(outp ? -15'h0C00 : 15'h0C00); // Quantization error
											// Full scale at +-150% to avoid excess noise
always @(posedge clk) begin		// Register update
	e1<=err;
	e2<=e1;
	out<=outp;
end

endmodule

