/////////////////////////////////////////////////
// Fast SPI controller
//  Jesús Arias (2023), public domain code
/////////////////////////////////////////////////
//  - SCK = clk/1 in fast mode
//  - SCK = clk/256 in slow mode
//  - 8 bit if wr8 strobe pulsed
//  - 16 bit if wr16 strobe pulsed

module SPI16_fast(
	input clk,
	input wr16,		// 16-bit write strobe
	input wr8,		// 8-bit write strobe
	input fast,		// control: SCK = clk or clk/256
	input [15:0]d,  // data to send (16 or 8 bits)
	output busy,	// flag, 1 = not yer done
	output [15:0]q,	// received data
	output sck,		// SPI master clk
	output mosi,	// SPI master output
	input  miso     // SPI master input
);

wire wr = wr16|wr8;
reg sz8=0;				// remember last write size
always @(posedge clk) if (wr) sz8<=wr8;

reg [7:0]divi=0;		// clock divider for slow mode
always @(posedge clk) divi<=wr ? 0 : divi+1;
wire dtc = (&divi) |fast;	// terminal count (always active in fast mode)

reg [4:0]bcnt='h1F;		// bit count (0 to 16)
reg [15:0]sh=0;			// shift register
assign busy=~bcnt[4];	

always @(posedge clk) begin
	bcnt<= wr ? (wr8 ? 8 : 0) : (busy&dtc ? bcnt+1 : bcnt);
	sh <= wr ? d : (busy&dtc ? {sh[14:0],smiso} : sh );
end

assign sck = busy & (fast ? (~clk) : divi[7]);
assign mosi = sz8 ? sh[7] : sh[15];
assign q= sz8 ? {8'h00,sh[7:0]}: sh;

reg smiso=0;	// sampled at SCK rising edges
always @(posedge sck) smiso <=miso;

endmodule

