//--------------------------------------------------------------------
// Sistema con CPU, RAM, UART, SPI, y DAC sigma-delta stereo
// J. Arias (2024), Public domain source
//--------------------------------------------------------------------

//`define INTERPOLATOR

`include "cpuV6.v"
`include "uart_simple.v"
`include "sd_dac.v"
`include "spi16f.v"

module SYSTEM (
	input clkin,	// 48MHz
	input reset,	// active High
	
	// UART
	output txd,
	input rxd,
	
	// SPI
	output sck,
	output mosi,
	output cs,
	input  miso,
	
	// Sigma delta
	output sdout1,
	output sdout2,
	
	// other
	output [4:0]leds
);

/////////////////////////////// Clocks ////////////////////////////////
reg clk=0;
reg [4:0]ckdiv=0;
always @(posedge clkin) begin
	clk<=~clk;						// main clock: x 1/2 = 24 MHz
	ckdiv<=ckdiv[4] ? 0 : ckdiv+1;	// DAC clock: x 1/17 = 2.82MHz
end
wire clk2M8 = ckdiv[4];

/////////////////////////////// MEMORY ////////////////////////////////

wire [15:0]cdi;	// CPU Data bus, Input
wire [15:0]cdo;	// CPU Data bus, Output
wire [15:0]ca;	// CPU Address bus

reg [15:0]mdo;	// Memory data output
parameter AW=11;
reg [15:0]ram[0:(1<<AW)-1];
initial
`ifndef SINT
	$readmemh("ROMgus.hex", ram);
`else
	$readmemh("rand.hex", ram);
`endif

always @(negedge clk) mdo<=ram[ca[AW-1:0]];
always @(posedge clk) if ((~iocs) & we) ram[ca[AW-1:0]]<=cdo;

//////////////////////////////////////////////////
//////					CPU					//////
//////////////////////////////////////////////////

assign cdi = iocs ? iodo : mdo;

wire we;

CPU_GUSV6 GUScpu0( .clk(clk), .we(we), .ca(ca), .cdo(cdo), .cdi(cdi), 
				  .reset(reset), .irq(irq), .ivector(ivector) );

//////////////////// IRQS ////////////////////
wire [2:0]ivector=3'b001;
wire irq;

// IRQEN reg
reg irqen=0;	
always @(posedge clk or posedge reset ) 
	if (reset) irqen<=0;
	else irqen <= iowe0 ? cdo[0] : irqen;
	
assign irq= dacirq & irqen;

/////////////////////////////////////////////////////////
///////////////////// Peripherals ///////////////////////
/////////////////////////////////////////////////////////

wire iocs=(ca[15:5]==11'b00000000001);	// IO select ($0020 to $003F)

// Write strobes
wire iowe0=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&(~ca[1])&(~ca[0]);
// wire iowe1=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&(~ca[1])&( ca[0]);
wire iowe2=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&(~ca[0]);
wire iowe3=(we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&( ca[0]);
wire iowe4=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&(~ca[0]);
wire iowe5=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&( ca[0]);
wire iowe6=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&( ca[1])&(~ca[0]);
wire iowe7=(we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&( ca[1])&( ca[0]);

// Read strobes
wire iore2=(~we)&(iocs)&(~ca[4])&(~ca[3])&(~ca[2])&( ca[1])&(~ca[0]);
//wire iore5=(~we)&(iocs)&(~ca[4])&(~ca[3])&( ca[2])&(~ca[1])&( ca[0]);

/////////////////////////////
// I/O data to CPU MUX
reg [15:0]iodo;
always@*
  case (ca[2:0])
     0 : iodo <=  {15'h0,irqen};	// IRQ Enable
     1 : iodo <=  {10'h0,sclk1,spibusy,uflags}; // Peripheral FLAGS
     2 : iodo <=  { 8'h0,urxdata};	// UART data
     3 : iodo <=  { 8'h0,gpout};	// GPOUT (as input)
     4 : iodo <=  spi_q;			// SPI redeived data
     default : iodo <= 16'hxxxx;
  endcase
  
/////////////////////////////
// Peripherals  

// GPOUT register: control + leds
reg [7:0]gpout=8'hff;
always @(posedge clk) if (iowe3) gpout<=cdo[7:0];
assign leds = gpout[4:0];
assign cs   = gpout[6];
assign spifast=gpout[7];

// UART
wire [7:0]urxdata;
wire [3:0]uflags;

UART_core #(.DIVIDER(208)) uart0 
	( .clk(clk), .d(cdo[7:0]), .wr(iowe2), .rd(iore2), .q(urxdata),
		.rxvalid(uflags[0]), .txrdy(uflags[1]),
		.rxoverr(uflags[2]), .rxframeer(uflags[3]),
		.nstop(1'b0), 
		.txd(txd), .rxd(rxd)
	);

// SPI
wire spibusy,spifast;
wire [15:0]spi_q;
wire sck,mosi,miso;

SPI16_fast spi0( .clk(clk), .wr16(iowe4), .wr8(iowe5), .d(cdo), .busy(spibusy),
				 .q(spi_q), .fast(spifast), .sck(sck), .mosi(mosi), .miso(miso)
);

// DACs
reg [11:0]dacreg1;
reg [11:0]dacreg2;
always @(posedge clk) begin
	if (iowe6) dacreg1<=cdo[15:4];
	if (iowe7) dacreg2<=cdo[15:4];
end

wire sclk1,sclk2;
sd_DAC dac1 ( .clk(clk2M8), .in(dacreg1), .out(sdout1), .sclk(sclk1) );
sd_DAC dac2 ( .clk(clk2M8), .in(dacreg2), .out(sdout2), .sclk(sclk2) );

// IRQ DAC
reg [1:0]osk=0;	// for falling edge detection
reg dacirq=0;
always @(posedge clk) begin
	osk<={osk[0],sclk1};
	dacirq<= (osk[1]&(~osk[0])) ? 1 : (iowe7 ? 0 : dacirq);
end

endmodule


